{$IFNDEF LINUX}
  {$IFDEF UNIX}
    {$WARNING 'This code might not work well on your version of unix.'}
  {$ELSE}
     {$FATAL 'This code was design to work with Linux based systems only.'}
  {$ENDIF}
{$ENDIF}
{$mode fpc} {$h+}
program sfplay;

uses SysUtils, ctypes, sndfile, BaseUnix;

Const
  BUFFER_LEN = 2048;

type
  TIntBuffer = array[0..BUFFER_LEN -1] of smallint;

Const
  { Values obtained from a C program - These are C macros -> soundcards.h }
  AFMT_S16_LE              = $00000010; { Little endian signed 16 }
  AFMT_S16_BE              = $00000020; { Big endian signed 16 }
  
  SOUND_PCM_WRITE_CHANNELS = -1073459194;
  SOUND_PCM_SETFMT         = -1073459195;
  SNDCTL_DSP_STEREO        = -1073459197;
  SOUND_PCM_WRITE_RATE     = -1073459198;
  SNDCTL_DSP_RESET         = 20480;
  SNDCTL_DSP_SYNC          = 20481;
  SNDCTL_DSP_POST          = 20488;

type
  TFileAccess    = (fasRead, fasWrite, fasExecute);
  TFileAccessSet = set of TFileAccess;

ResourceString
  SPlaying            = 'Playing : ';
  SErrChannels        = 'Error : Number of channels not supported: ';
  SErrOpeningDevice   = 'Could not open sound device';
  SErrNoSoundDevice   = 'Could not find a sound device';
  SErrSettingStereo   = 'Could not set stereo';
  SErrResettingDevice = 'Could not reset DSP device';
  SErrSetWriteBits    = 'Could not set write bits to 16';
  SErrSetChannels     = 'Could not set channels';
  SErrSetSampleRate   = 'Could not set sync mode';
  SErrSetSyncMode     = 'Could not set sync mode';
  SErrSetFormat       = 'Could not set format';
  SErrPost            = 'Could not set post';
  SErrFileUnreadable  = 'The user running the program do not have permmisions to read the.';
  SErrNoFileWasGiven  = 'Error: No audio file was given.';

function ioctl  (d : cint; request : cint; a : pointer) : cint; cdecl; external 'c' name 'ioctl';
function usleep (__useconds : culong)                   : cint; cdecl; external 'c' name 'usleep';

procedure DisplayError(Msg : String);
begin
  writeln(StdErr, Msg);
end;

function FileExistsAndAccessible(const FileName : String; FileAccess : TFileAccessSet) : Boolean;
var
  Status     : Boolean;
begin
  Status := fpAccess(FileName, F_OK) = 0;

  if fasRead in FileAccess then
    begin
      Status := Status and (fpAccess(FileName, R_OK) = 0);
    end;

  if fasWrite in FileAccess then
    begin
      Status := Status and (fpAccess(FileName, W_OK) = 0);
    end;

  if fasExecute in FileAccess then
    begin
      Status := Status and (fpAccess(FileName, X_OK) = 0);
    end;

  FileExistsAndAccessible := Status;
end;

Function OpenDSPDevice(Channels, Samplerate : LongInt; var Descriptior : Longint) : Boolean;
var
 fd, stereo, fmt : Longint;

begin
  if FileExistsAndAccessible('/dev/dsp', [fasWrite]) then
    fd:=fpOpen('/dev/dsp', O_WRONLY, 0)
  else begin
         DisplayError(SErrNoSoundDevice);
         OpenDSPDevice := False;
         exit;
       end;

  if fd < 0 then
    begin
      DisplayError(SErrOpeningDevice);
      OpenDSPDevice := False;
      exit;
    end;
  
  Stereo := 0;
  if IOCtl(fd, SNDCTL_DSP_STEREO  , @stereo) <> 0 then
    begin
      DisplayError(SErrSettingStereo);
      OpenDSPDevice := False;
      exit;
    end;

  if IOCtl (fd, SNDCTL_DSP_RESET, Nil) <> 0 then
    begin
      DisplayError(SErrResettingDevice);
      OpenDSPDevice := False;
      exit;
    end;

  fmt := {$IFDEF ENDIAN_LITTLE} AFMT_S16_LE {$ELSE} AFMT_S16_BE {$ENDIF};
  if IOCtl(fd, SOUND_PCM_SETFMT, @fmt) <> 0 then
    begin
      DisplayError(SErrSetFormat);
      OpenDSPDevice := False;
      exit;
    end;

  if IOCtl (fd, SOUND_PCM_WRITE_CHANNELS, @channels) <> 0 then
    begin
      DisplayError(SErrSetChannels);
      OpenDSPDevice := False;
      exit;
    end;

  if IOCtl (fd, SOUND_PCM_WRITE_RATE, @SampleRate) <> 0 then
    begin
      DisplayError(SErrSetSampleRate);
      OpenDSPDevice := False;
      exit;
    end;

  if IOCtl (fd, SNDCTL_DSP_SYNC, Nil) <> 0 then
    begin
      DisplayError(SErrSetSyncMode);
      OpenDSPDevice := False;
      exit;
    end;

  Descriptior   := fd;
  OpenDSPDevice := true;
end;

procedure PlayFile (FileName : String);
var
  Buffer        : TIntBuffer;
  FloatBuffer   : Array[0..BUFFER_LEN -1] of single;
  SoundFile     : PSndFile;
  Info          : TSF_INFO;
  m,
  audio_device,
  subformat     : Longint;
  scale         : double;
  readcount     : cshort;
  delay_time    : culong;
  
  procedure write_buffer; inline;
  var
   start_count,
   output_count,
   write_count,
   max_write     : clong;
  begin
    start_count  := 0;
    output_count := readcount * sizeof(cshort);
    max_write    := 0;
    //writeln('output_count : ', output_count);

    while (output_count > 0) do // write as much data as possible
      begin
        write_count := fpwrite (audio_device, Buffer[start_count], output_count);
        // Give the audio output time to catch up.
        //writeln('write_buffer -> write_count : [',write_count, ']');
        inc (max_write, write_count);
        if (write_count > 0) then
          begin
            dec (output_count, write_count);
            inc (start_count, write_count);
            //writeln('output_count : ', output_count);
            //writeln('start_count : ', start_count);
          end
		 	  else begin
			  		   // Give the audio output time to catch up.
				  	   usleep (delay_time);
             end;
      end;

    //writeln('max_write : [',max_write, ']');
  end;
  
begin
  if not FileExistsAndAccessible(FileName, [fasRead]) then
    begin
      DisplayError(SErrFileUnreadable);
      exit;
    end;
  FillChar (Info, SizeOf(Info), 0);
  
  writeln(SPlaying, FileName);
  SoundFile := nil;
  SoundFile := sf_open(pcChar(FileName), SFM_READ, @Info);
  if SoundFile = nil then
    begin
      sf_perror(Nil);
      exit;
    end;

  delay_time := (BUFFER_LEN * 1000000) div (Info.samplerate * 4) ;
  //writeln('delay_time : [', delay_time, ']');
    
  //writeln('Channels : [',Info.channels, ']');
  if (Info.channels < 1) or (Info.channels > 2) then
    begin
      DisplayError(SErrChannels);
      sf_close(SoundFile);
      exit;
    end;
    
   if not OpenDSPDevice(Info.channels, Info.samplerate, audio_device) then
     begin
       sf_close(SoundFile);
       halt(1);
     end;
     
  subformat    := Info.format and SF_FORMAT_SUBMASK;
  if (subformat = SF_FORMAT_FLOAT) or (subformat = SF_FORMAT_DOUBLE) then
    begin
      sf_command(SoundFile, SFC_CALC_SIGNAL_MAX, @scale, sizeof(scale));
      
      //writeln('Scale : [', scale, ']');
      
      if scale < 1e-10 then
        scale := 1.0
      else
        scale := 32700.0 / scale;
        
      //writeln('scale : [', scale, ']');

      readcount := sf_read_float (SoundFile, @FloatBuffer[0], BUFFER_LEN);
      while (readcount <> 0) do
        //writeln('sf_read_float readcount : [', readcount, ']');
        begin
          for m := 0 to readcount - 1 do
            begin
              buffer[m] := round(scale * FloatBuffer [m]);
            end;
          write_buffer;
          readcount := sf_read_float (SoundFile, @FloatBuffer[0], BUFFER_LEN );
        end;
    end
  else begin
         readcount := sf_read_short (SoundFile, @Buffer[0], BUFFER_LEN);
         while (readcount <> 0) do
           begin
             //writeln('sf_read_short readcount : [', readcount, ']');
             write_buffer;
             readcount := sf_read_short (SoundFile, @Buffer[0], BUFFER_LEN );
           end;
           
       end;

  if IOCtl (audio_device, SNDCTL_DSP_POST, PByte(0)) = -1 then
    DisplayError(SErrPost);
    
  if IOCtl(audio_device, SNDCTL_DSP_SYNC, PByte(0)) = -1 then
    DisplayError(SErrSetSyncMode);
    
  FpClose(audio_device);
  sf_close(SoundFile);
end;

Var
  I : Integer;

begin
  if paramcount < 1 then
    begin
      writeln(STDERR, SErrNoFileWasGiven);
      writeln(STDERR,'Usage : ', ExtractFileName(ParamStr(0)), ' <audio file> [audio file ...]');
      writeln(STDERR, '');
      exit;
    end;

  For I:=1 to ParamCount do
    PlayFile(Paramstr(i));
end.
